/*3456789_123456789_123456789_123456789_123456789_123456789_123456789_123456789_


	v_vecmat.cpp:  (portion of a) homebrewed vector/matrix library; includes
		2D, 3D, 4D vector and matrix classes, and an arbitrary-sized matrix.

 Most functions use C++ pass-by-reference for input parameters, and all use local 
temporary storage of intermediate results so that in-place calculations never
cause problems (e.g. any object can invoke a member function using itself as one 
or more of the arguments).  The results are placed in the calling object, unless 
the result is of a different type; then results are the returned value.  For 
example, the Mat3 matrix class has a member function named mm_mpy() that 
multiplies a 'left' and 'rt' matrix.  Given three Mat3 objects named mmA,mmB, and 
mmC, use the statement
	 	mmC.mm_mpy(mmA,mmB);  // matrix multiply: mmC = mmA * mmB 
You may make in-place calculations like this:
		mmB.mm_mpy(mmB,mmB);	// new mmB = mmB * mmB
without error, because the temporary storage within each function ensures inputs 
and outputs are not intermixed.

The arbitrary-size matrix class 'matrx' was written for compatibility with 
"Numerical Recipes in (C|C++|FORTRAN):The Art of Scientific Computing" by
William H. Press, Saul A. Teukolsky,William T. Vetterling, Brian P. Flannery
Cambridge Univ. Press.(see http://www.nr.com/).  All 'matrx' function members 
have two versions, such as put0(), put1() or get0(),get1().  These '0' suffixed
member functions use zero-indexed addressing (C-style), and the '1' suffixed
members use 1-indexed addressing (FORTRAN style). 
Note that ALL functions have the 0 or 1 suffix--this helps prevent accidental
use of the wrong function (a VERY nasty bug to find).

REVISIONS:
5/14/93		J. Tumblin	CREATED from my3vec.c
9?/??/98	J. Tumblin added separate integer matrix/vector classes Ivec,Imat
12/12/98	J. Tumblin added arbitrary matrix class (from old mtx,matrx)
12/14/04	J. Tumblin cruft removal, commenting for 'Quadrilateral Int.' paper 
==============================================================================*/
#if !defined(__JT__V_VECMAT)// if we haven't already included this file,
#define __JT__V_VECMAT		// compile it! (#endif at end-of-file).

//==================
//  DEFINE stmts;
//==================
#define VECTYPE double			// 'float' or 'double'-- sets element data type.
#define IS_FLOAT FALSE			// define matrix/vector type (double or float)
#define IS_DOUBLE TRUE			

#if !defined(M_PI)
	#define M_PI 3.1415926535897932384626433832795  // (from MS calculator app)
								// Because VC++ doesn't have M_PI constant 
#endif

#if !defined(BOOL)				// Just in case your compiler doesn't have 
#define BOOL short int			// a 'boolean' type.
#endif

#if !defined(TRUE)
#define TRUE 1
#endif

#if !defined(FALSE)
#define FALSE 0
#endif

#define DPF printf				// for non-MS Windows debugging,

//==================
// FWD declarations;
//==================
class Mat2;						// Do fwd-declare of Mat2: Vec2 uses it.
class Mat3;						// DO fwd-...
class Mat4;						// DO fwd-...
class Matrx;					// DO fwd-...

//==============================================================================
//================================                ==============================
//================================   2x2 Vectors  ==============================
//================================                ==============================
//==============================================================================

class Vec2 {
//==============================================================================
// a 1-D homogeneous 'column vector';
//					[row[0] ]
//					[row[1] ]												  
								//-----------------------DATA-------------------
    public:
	VECTYPE row[2];	
								//======-----------------FCNS-------------------
    public:
	Vec2() {								// void constructor
		row[0] = (VECTYPE)0.0; row[1] = (VECTYPE)0.0; };
	
	Vec2(VECTYPE r0,VECTYPE r1=0) {			// explicit constructor
	    row[0] = r0; row[1] = r1; };

    void put(VECTYPE r0,VECTYPE r1) {		// set vec elements 
		row[0] = r0; row[1] = r1; };

    VECTYPE get0(int rowNum) {				// get vec elements 
		return row[rowNum]; };

	void zero() {							// clear a vector to 0.0
	    row[0] = (VECTYPE)0.0; row[1] = (VECTYPE)0.0; };

	void prnt();							// prettyprint a vec
	void prnt(char* name);					// print a labeled vec
	VECTYPE norm(Vec2& src);				// normalize, & return length
	void add(Vec2& src1, Vec2& src2);		// add: [this] = src1+src2
	void addIn(Vec2& src);					// accumulate: [this] += src;
	void sub(Vec2& src1, Vec2& src2);		// subtract;[this] = src1-src2
	void scale(Vec2& src1, VECTYPE c);		// scale: [this]= src1.*c
	VECTYPE dot(Vec2& src1, Vec2& src2);	// dot product; retn [in1].[in2]
	VECTYPE cross(Vec2& src1);				// 2D cross product; [this]X[in1]
	void mpy(Mat2& srcM, Vec2& srcV);		// matrix-vector mpy:
											// [this] = [in1m][in2v]
// Form Quadrilateral Integration:
	BOOL isCloserTo(Vec2& srcA, Vec2& srcB);// TRUE iff srcA is closer than srcB.
	VECTYPE CCWAngleTo(Vec2& dest);			// Find angle in CCW radians,
											// (0-to-2pi) from 'this' to 'dest'.
	VECTYPE signedAngleTo(Vec2& dest);		// Get angle in radians (+/- PI)
											// from 'this' to 'dest' vector.
	BOOL isInsideQuad(Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3);
											// TRUE iff 'this' is a point
											// enclosed by quadrilateral p0-p3
	int classifyQuad(Vec2& p0, Vec2& p1, Vec2& p2, Vec2 & p3);
											// Is it convex? back-facing? triangle?
											// chevron? bow-tie? degenerate?
	void testClassifyQuad(void);			// Generates test quads, classifies.
};

class Mat2 { 	
//==============================================================================
// A 1D homogeneous matrix data type.  Note colrow[][] indices are 
//		   just like colrow[x][y] (x,y position within matrix).
//		   [ col[0][0]  colrow[1][0] ]
//		   [ col[0][1]  colrow[1][1] ]
//		   [ col[0][2]  colrow[1][2] ]									

							//-----------------DATA-------------------
    public:
	VECTYPE colrow[2][2]; 				// gives [column#][row#], equivalent to
										// to the x,y position in matrix
							//-----------------FCNS-------------------
    public:
	Mat2(void);							// void constructor
	void zero(void);					// clear matrix to 0.0
	void ident(void);					// set to identity matrix
	void scale(Mat2& in1, VECTYPE scale);// make scaled copy of a matrix
	void prnt(void);					// prettyprint matrix
	void prnt(char *name);				// print a labeled matrix

	void mpy(Mat2& left, Mat2& rt);	
										// matrix-matrix mpy:[this] = [left][rt]
	void transpose(void);				// in-place transpose

	//--------------------For 2D Vector operations:
//	void makeRot90();					// CCW 90 degree rotation.
//	void makeRot180();					// CCW 180 degree rotation.
	void makeRot270();					// CCW 270 degree rotation.	

	//--------------------Compatibility with other matrix classes
	void copyTo(Matrx &out);			// resize out, copy contents to it.

};

//==============================================================================
//================================                ==============================
//================================   3x3 Vectors  ==============================
//================================                ==============================
//==============================================================================

class Vec3 {
//==============================================================================
// a 2-D homogeneous 'column vector';
//					[row[0] ]
//					[row[1] ]
//					[row[2] ]										  

							//-----------------DATA-------------------
    public:
	VECTYPE row[3];	
							//-----------------FCNS-------------------
    public:
	Vec3(void) {						// void constructor
		row[0] = (VECTYPE)0.0; row[1] = (VECTYPE)0.0; row[2] = (VECTYPE)0.0; };

	Vec3(VECTYPE r0,VECTYPE r1,VECTYPE r2) {
									// explicit constructor
		row[0] = r0; row[1] = r1; row[2] = r2; };
	void put(VECTYPE r0,VECTYPE r1,VECTYPE r2) {	// set vec elements 
		row[0] = r0; row[1] = r1; row[2] = r2; };
	void zero(void) {					// clear a vector to 0.0
	    row[0] = (VECTYPE)0.0; row[1] = (VECTYPE)0.0; row[2] = (VECTYPE)0.0; };
	void prnt(void);					// prettyprint a vec
	void prnt(char* name);				// print a labeled vec
	VECTYPE norm(Vec3& src);			// normalize, & return length
	void add(Vec3& src1, Vec3& src2);	// add: [this] = src1+src2
	void sub(Vec3& src1, Vec3& src2);	// subtract;[this] = src1-src2
	void scale(Vec3& src1, VECTYPE c);	// scale: [this]= src1.*c
	VECTYPE dot(Vec3& src1, Vec3& src2);// dot product; retn [in1].[in2]
	void cross(Vec3& src1, Vec3& src2);	// cross product; [this]= [in1]X[in2]
	void mpy(Mat3& srcM, Vec3& srcV);	// matrix-vector mpy:
										//				  [this] = [in1m][in2v]
};

class Mat3 { 	
//==============================================================================
// A 2D homogeneous matrix data type.  Note colrow[][] indices are 
//		   just like colrow[x][y] (x,y position within matrix).
//		   [ col[0][0]  colrow[1][0]  colrow[2][0] ]
//		   [ col[0][1]  colrow[1][1]  colrow[2][1] ]
//		   [ col[0][2]  colrow[1][2]  colrow[2][2] ]				  

							//-----------------DATA-------------------
    public:
	VECTYPE colrow[3][3]; 				// gives [column#][row#], equiv
										// to x,y position in matrix,
										// and allows 'a=b' assignments.
							//-----------------FCNS-------------------
    public:
	Mat3(void) { zero(); };				// void constructor
//	Vec3 getcol(int col);				// return a column.
//	void putcol(Vec3& in1, int col);	// overwrite a column.
	void zero(void);					// clear matrix to 0.0
	void ident(void);					// set to identity matrix
	void copy(Mat3& src);				// element-by-element copy of matrix.
	void prnt(void);					// prettyprint matrix
	void prnt(char *name);				// print a labeled matrix
	void mpy(Mat3& left, Mat3& rt);		// matrix-matrix mpy:[this] = [left][rt]
	void convolve(Mat2& inA2, Mat2& inB2); // for 2D polynomial multiply
	void transpose(void);				// in-place transpose

	//--------------------Compatibility with other matrix classes
	void copyTo(Matrx &out);			// resize out, copy contents to it.
};

//==============================================================================
//================================                ==============================
//================================   4x4 Vectors  ==============================
//================================                ==============================
//==============================================================================

class Vec4 {			
//==============================================================================
// an 3-D homogeneous 'column vector';
//					[row[0] ]
//					[row[1] ]
//					[row[2] ]
//					[row[3] ]										  

							//-----------------DATA-------------------
    public:
	VECTYPE row[4];	
							//-----------------FCNS-------------------
    public:
	Vec4(void) { zero(); };				// void constructor
	Vec4(VECTYPE r0,VECTYPE r1,VECTYPE r2,VECTYPE r3) {
										// explicit constructor
		    row[0] = r0; row[1] = r1; row[2] = r2; row[3] = r3; };
	void put(VECTYPE r0,VECTYPE r1,VECTYPE r2,VECTYPE r3) {// set vec elements 
		    row[0] = r0; row[1] = r1; row[2] = r2; row[3] = r3; };
	void zero(void) {					// clear a vector to 0.0
	    row[0] = (VECTYPE)0.0; row[1] = (VECTYPE)0.0; 
		row[2] = (VECTYPE)0.0; row[3] = (VECTYPE)0.0; };
	void prnt(void);					// prettyprint a vec
	void prnt(char* name);				// print a labeled vec
	VECTYPE norm(Vec4& src); 			// normalize, then return length
	void add(Vec4& src1, Vec4& src2);  	// add: [this] = src1+src2
	void sub(Vec4& src1, Vec4& src2);	// subtract;[this] = src1-src2
	void scale(Vec4& src1, VECTYPE c);	// scale: [this]= src1.*c
	VECTYPE dot(Vec4& src1, Vec4& src2);// dot product; retn [in1].[in2]
	void cross(Vec4& src1, Vec4& src2);	// homogeneous coord cross prod;
										// [this]=[src1]X[src2]
	void mpy(Mat4& srcM, Vec4& srcV);	// matrix-vector mpy:
										// [this] = [in1m][in2v]
	//--------------------Compatibility with other matrix classes
	void copyTo(Matrx &out);			// resize out, copy contents to it.
};

class Mat4 {
//==============================================================================
// matrix data type for 3D homogeneous matrices; 
// NOTE colrow[][] indices are colrow[x][y] (x,y position within matrix).
//
//		   [ col[0][0]  colrow[1][0]  colrow[2][0] ... ]
//		   [ col[0][1]  colrow[1][1]  colrow[2][1] ... ]
//		   [ col[0][2]  colrow[1][2]  colrow[2][2] ... ]
//		   [   ...		...			  ...		   ... ]
//
							//-----------------DATA-------------------
    public:
	VECTYPE colrow[4][4]; 		// gives [column#][row#], equiv to x,y 
								// position in matrix
							//-----------------FCNS-------------------
    public:
	Mat4(void) { zero(); };				// void constructor
	void zero(void);					// clear matrix to 0.0
	void ident(void);					// set to identity matrix
	void prnt(void);					// prettyprint matrix
	void prnt(char *name);				// print a labeled matrix
	void mpy(Mat4& left, Mat4& rt);		// matrix-matrix mpy:
										// [this] = [this][rt]
	void convolve(Mat2& inA2, Mat3& inB3); // for 2D polynomial multiply
	void transpose(void);				// in-place transpose
	//--------------------Compatibility with other matrix classes
	void copyTo(Matrx &out);			// resize out, copy contents to it.

};
 
//==============================================================================
//================================                ==============================
//================================   Arbitrary    ==============================
//================================  Sized Matrix  ==============================
//================================                ==============================
//==============================================================================


class Matrx {
//==============================================================================
// An arbitrary-sized matrix class using a storage method compatible with IMSL 
// libraries (linkable FORTRAN).  PLEASE PLEASE PLEASE don't directly access 
// the 'a' array; use the 'get()' and 'put()' member functions, because I may 
// change the matrix storage methods later to add 'sparse' matrix types, etc.
// Written for use with the usual 0-based array indices of C,C++, BUT you can 
// use the 1-based indices common to FORTRAN code and all the "Numerical Recipes
// in C" code by using 'get1()' and put1()' instead of the zero-based 'get0()' 
// and 'put0()'.
//	12/14/98: added dyn.alloc'd array of pointers to each row so we are
//				also compatible with Numerical Recipes code.

//private:			    			//-----------------DATA-----------------
public:
    VECTYPE *a;			   			// dynam. allocated space at this ptr holds 
									// 2D array of matrix
									// entries in row-major order--
									// A00,A10,A20,.. (all of col 0,
									// then all of col 1, etc)
									// stored value is element
    int xsize;						// # columns in the matrix.
	int ysize;						// # of rows

public:			    			//-----------------Access-----------------
    Matrx(void);					// make an 'empty' object
    ~Matrx(void);					// destructor; de-alloc memory
    BOOL wipe(void);				// release space for matrix.
    void sizer(int iysize,int ixsize);	// allocate space for matrix;
									// sizer(#rows,#columns)
	void sizer(Matrx& src);	

	int getXsize(void) { return(xsize);};
	int getYsize(void) { return(ysize);};		 
    BOOL copy(Matrx& src);			// auto-resizing entry-by-entry copy
    								// (no change to our object's name) 
	void copy0Col(int destCol, Matrx &src, int srcCol);
									// copy one column from 'src' matrix.
	void put0(int row,int col, VECTYPE val);
	VECTYPE get0(int row, int col);	// read/write 'val' at  row,col.	
									// (uses zero-based indexing)
	void put1(int row,int col, VECTYPE val);
	VECTYPE get1(int row, int col);	// read/write 'val' at  row,col.	
									// (uses ONE-based indexing)
	void addTo0(int row, int col, VECTYPE val);
									// add 'val' to current value at row,col.
									// (uses zero-based indexing)
	void addTo1(int row, int col, VECTYPE val);
									// add 'val' to current value at row,col.
									// (uses ONE-based indexing)
									// Write values at (r0,col),(r1,col),...
	VECTYPE* getPtr0(int row=0, int col=0) { //(zero-based indexing)		
									// return pointer to element;
									// default to 1st element.
		return(&(a[row + col*ysize])); };
	VECTYPE* getPtr1(int row=1, int col=1) {// (one-based indexing)
		return(&(a[(row-1) + (col-1)*ysize])); };
								//---------------MATH-----------------------
	void mpyScalar(VECTYPE wt);		// scale every element by 'wt'.
	BOOL mmpy(Matrx& left, Matrx& rt);// replace our current contents with 
									// result of matrix multiply [left][rt]. 
									// Return FALSE on err.
	BOOL add(Matrx& left,Matrx& rt);// 'this'=left+rt('this' as input is OK)
	BOOL sub(Matrx& left,Matrx& rt);// 'this'=left-rt('this' as input is OK)
	BOOL transpose(Matrx& src);		// in-place is OK(src=*this)

								//---------------------------------------
								// 2D Polynomial Forms: describe any polynomial
									// of 2 variables (a 'poly2') as a matrix:
									// f(u,v) = [V]*[C]*[U], where 
									//  V = [1, v, v^2, v^3, ...]
									//  U = [1, u, u^2, u^3, ...]^T
									//  C = coefficient matrix.

								//---------------------------------------
								// 2D Polynomial Forms: describe any polynomial
									// of 2 variables (a 'poly2') as a matrix:
									// f(u,v) = [V]*[C]*[U], where 
									//  V = [1, v, v^2, v^3, ...]
									//  U = [1, u, u^2, u^3, ...]^T
									//  C = coefficient matrix.
	VECTYPE poly2Eval(VECTYPE u0, VECTYPE v0);
									// Evaluate the poly2 described by 'this'
									// at the point (u0,v0).
	void convolve(Mat3& inA3, Mat3& inB3);
	void convolve(Matrx& inA, Matrx& inB);
									// same, but reversed-order arguments:
	void makeF3_hat(Mat3& F3, Mat2& Bx, Mat2& By);
									// For Bixels; use bi-quadratic function F3
									// and bilinear maps Bx, By to construct the
									// coef. matrix F3hat for f(x(u,v),y(u,v));
	void poly2Pow(int pow, Matrx& inA);
									// Raise a poly2 to a power; self-convolve
									// the coeff. matrix inA for (pow-1) times.
	VECTYPE poly2quadIntegrateF3(Mat3& F3, 
								   Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3);									
									// compute integral over a quadrilateral 
									// region defined by corner points p0--p3.
									// (demo for JGT 2004 paper submission)
	VECTYPE poly2quadMC(Matrx& F3, Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3);
									// Monte-Carlo soln. to poly2quadIntegrateF3()
									// used for testing/verifying the former.
	void poly2convexQuadTestPrnt(void);	 // Test poly2quadIntegrateF3() against
									// poly2quadMC() on random CONVEX quads for 
									// random cubic polynomials; print results.
	void poly2arbitraryQuadTestPrnt(void);	 // Test poly2quadIntegrateF3() against
									// poly2quadMC() on random ARBITRARY quads for 
									// random cubic polynomials; print results.

								// print results.
								//--------------------------------------
	void randFill(VECTYPE min=(VECTYPE)0.0,VECTYPE max=(VECTYPE)1.0);
									// fill matrix with random #s
									// 		(default to 0.0--1.0 range)
    void zero(void);				// fill matrix with zeroes
	void ident(void);				// make identity matrix.
	void countR(void);				// fill with counting; row-major
	void countC(void);				// fill with counting; column-major
	BOOL isEqual(Matrx &src,VECTYPE tol2=1.0E-15); 
									// TRUE iff 'this'=='src', (+/- tol)
	void copyTo(Mat2& out);			// Copy first 2x2 elements to 'out'
	void copyTo(Mat3& out);			// Copy first 3x3 elements to 'out'
	void copyTo(Mat4& out);			// Copy first 4x4 elements to 'out'
	void textOut(FILE* fptr, char* title, int linewidth);		
									// make printout in stream 'fptr'
	void prnt(const char *name);	// Print to debug window using 'TRACE'.
									// usage: mymat.prnt("mymat");
};

#undef PIXTYPE						// Locals only!

#endif		// conditional compile (#if defined _V_VECMAT at top of file).
